#![feature(impl_trait_in_assoc_type)]

use std::{
    any::Any,
    collections::{hash_map, HashMap},
    fmt::Display,
    future::{ready, Future, Ready},
    marker::PhantomData,
    net::IpAddr,
    ops::DerefMut,
    pin::Pin,
    sync::Arc,
};

use btlib::{bterr, crypto::Creds, error::StringError, BlockPath, Result};
use btserde::{field_helpers::smart_ptr, from_slice, to_vec, write_to};
use bttp::{DeserCallback, MsgCallback, Receiver, Replier, Transmitter};
use serde::{de::DeserializeOwned, Deserialize, Serialize};
use tokio::{
    sync::{mpsc, oneshot, Mutex, RwLock},
    task::JoinHandle,
};
use uuid::Uuid;

/// Declares a new [Runtime] which listens for messages at the given IP address and uses the given
/// [Creds]. Runtimes are intended to be created once in a process's lifetime and continue running
/// until the process exits.
#[macro_export]
macro_rules! declare_runtime {
    ($name:ident, $ip_addr:expr, $creds:expr) => {
        ::lazy_static::lazy_static! {
            static ref $name: &'static ::btrun::Runtime = {
                ::lazy_static::lazy_static! {
                    static ref RUNTIME: ::btrun::Runtime =  ::btrun::Runtime::_new($creds).unwrap();
                    static ref RECEIVER: ::bttp::Receiver = _new_receiver($ip_addr, $creds, &*RUNTIME);
                }
                // By dereferencing RECEIVER we ensure it is started.
                let _ = &*RECEIVER;
                &*RUNTIME
            };
        }
    };
}

/// This function is not intended to be called by downstream crates.
#[doc(hidden)]
pub fn _new_receiver<C>(ip_addr: IpAddr, creds: Arc<C>, runtime: &'static Runtime) -> Receiver
where
    C: 'static + Send + Sync + Creds,
{
    let callback = RuntimeCallback::new(runtime);
    Receiver::new(ip_addr, creds, callback).unwrap()
}

/// Type used to implement an actor's mailbox.
pub type Mailbox<T> = mpsc::Receiver<Envelope<T>>;

/// An actor runtime.
///
/// Actors can be activated by the runtime and execute autonomously until they return. Running
/// actors can be sent messages using the `send` method, which does not wait for a response from the
/// recipient. If a reply is needed, then `call` can be used, which returns a future that will
/// be ready when the reply has been received.
pub struct Runtime {
    path: Arc<BlockPath>,
    handles: RwLock<HashMap<Uuid, ActorHandle>>,
    peers: RwLock<HashMap<Arc<BlockPath>, Transmitter>>,
    registry: RwLock<HashMap<ServiceId, ServiceRecord>>,
}

impl Runtime {
    ///  This method is not intended to be called directly by downstream crates. Use the macro
    /// [declare_runtime] to create a [Runtime].
    ///
    /// If you create a non-static [Runtime], your process will panic when it is dropped.
    #[doc(hidden)]
    pub fn _new<C: 'static + Send + Sync + Creds>(creds: Arc<C>) -> Result<Runtime> {
        let path = Arc::new(creds.bind_path()?);
        Ok(Runtime {
            path,
            handles: RwLock::new(HashMap::new()),
            peers: RwLock::new(HashMap::new()),
            registry: RwLock::new(HashMap::new()),
        })
    }

    pub fn path(&self) -> &Arc<BlockPath> {
        &self.path
    }

    /// Returns the number of actors that are currently executing in this [Runtime].
    pub async fn num_running(&self) -> usize {
        let guard = self.handles.read().await;
        guard.len()
    }

    /// Sends a message to the actor identified by the given [ActorName].
    pub async fn send<T: 'static + SendMsg>(
        &self,
        to: ActorName,
        from: ActorName,
        msg: T,
    ) -> Result<()> {
        if to.path == self.path {
            let guard = self.handles.read().await;
            if let Some(handle) = guard.get(&to.act_id) {
                handle.send(from, msg).await
            } else {
                Err(bterr!("invalid actor name"))
            }
        } else {
            let guard = self.peers.read().await;
            if let Some(peer) = guard.get(&to.path) {
                let buf = to_vec(&msg)?;
                let wire_msg = WireMsg {
                    to,
                    from,
                    payload: &buf,
                };
                peer.send(wire_msg).await
            } else {
                // TODO: Use the filesystem to discover the address of the recipient and connect to
                // it.
                todo!()
            }
        }
    }

    /// Sends a message to the service identified by [ServiceName].
    pub async fn send_service<T: 'static + SendMsg>(
        &'static self,
        to: ServiceAddr,
        from: ActorName,
        msg: T,
    ) -> Result<()> {
        if to.path().as_ref() == self.path.as_ref() {
            let actor_id = self.service_provider(&to).await?;
            let handles = self.handles.read().await;
            if let Some(handle) = handles.get(&actor_id) {
                handle.send(from, msg).await
            } else {
                panic!(
                    "Service record '{}' had a non-existent actor with ID '{}'.",
                    to.service_id(),
                    actor_id
                );
            }
        } else {
            todo!("Send the message to an appropriate peer.")
        }
    }

    /// Sends a message to the actor identified by the given [ActorName] and returns a future which
    /// is ready when a reply has been received.
    pub async fn call<T: 'static + CallMsg>(
        &self,
        to: ActorName,
        from: ActorName,
        msg: T,
    ) -> Result<T::Reply> {
        if to.path == self.path {
            let guard = self.handles.read().await;
            if let Some(handle) = guard.get(&to.act_id) {
                handle.call_through(msg).await
            } else {
                Err(bterr!("invalid actor name"))
            }
        } else {
            let guard = self.peers.read().await;
            if let Some(peer) = guard.get(&to.path) {
                let buf = to_vec(&msg)?;
                let wire_msg = WireMsg {
                    to,
                    from,
                    payload: &buf,
                };
                peer.call(wire_msg, ReplyCallback::<T>::new()).await?
            } else {
                todo!("Use the filesystem to find the address of the recipient and connect to it.")
            }
        }
    }

    /// Calls a service identified by [ServiceName].
    pub async fn call_service<T: 'static + CallMsg>(
        &'static self,
        to: ServiceAddr,
        msg: T,
    ) -> Result<T::Reply> {
        if to.path().as_ref() == self.path.as_ref() {
            let actor_id = self.service_provider(&to).await?;
            let handles = self.handles.read().await;
            if let Some(handle) = handles.get(&actor_id) {
                handle.call_through(msg).await
            } else {
                panic!(
                    "Service record '{}' had a non-existent actor with ID '{}'.",
                    to.service_id(),
                    actor_id
                );
            }
        } else {
            todo!("Send the message to an appropriate peer.")
        }
    }

    fn service_not_registered_err(id: &ServiceId) -> btlib::Error {
        bterr!("Service is not registered: '{id}'")
    }

    async fn service_provider(&'static self, to: &ServiceAddr) -> Result<Uuid> {
        let actor_id = {
            let registry = self.registry.read().await;
            if let Some(record) = registry.get(to.service_id()) {
                record.actor_ids.first().copied()
            } else {
                return Err(Self::service_not_registered_err(to.service_id()));
            }
        };
        let actor_id = if let Some(actor_id) = actor_id {
            actor_id
        } else {
            let mut registry = self.registry.write().await;
            if let Some(record) = registry.get_mut(to.service_id()) {
                // It's possible that another thread got the write lock before us and they
                // already spawned an actor.
                if record.actor_ids.is_empty() {
                    let spawner = record.spawner.as_ref();
                    let actor_name = spawner(self).await?;
                    let actor_id = actor_name.act_id;
                    record.actor_ids.push(actor_id);
                    actor_id
                } else {
                    record.actor_ids[0]
                }
            } else {
                return Err(Self::service_not_registered_err(to.service_id()));
            }
        };
        Ok(actor_id)
    }

    /// Spawns a new actor using the given activator function and returns a handle to it.
    pub async fn spawn<Msg, F, Fut>(&'static self, activator: F) -> ActorName
    where
        Msg: 'static + CallMsg,
        Fut: 'static + Send + Future<Output = ()>,
        F: FnOnce(&'static Runtime, Mailbox<Msg>, Uuid) -> Fut,
    {
        let mut guard = self.handles.write().await;
        let act_id = {
            let mut act_id = Uuid::new_v4();
            while guard.contains_key(&act_id) {
                act_id = Uuid::new_v4();
            }
            act_id
        };
        let act_name = self.actor_name(act_id);
        let (tx, rx) = mpsc::channel::<Envelope<Msg>>(MAILBOX_LIMIT);
        // The deliverer closure is responsible for deserializing messages received over the wire
        // and delivering them to the actor's mailbox, and sending replies to call messages.
        let deliverer = {
            let buffer = Arc::new(Mutex::new(Vec::<u8>::new()));
            let tx = tx.clone();
            let act_name = act_name.clone();
            move |envelope: WireEnvelope| {
                let (wire_msg, replier) = envelope.into_parts();
                let result = from_slice(wire_msg.payload);
                let buffer = buffer.clone();
                let tx = tx.clone();
                let act_name = act_name.clone();
                let fut: FutureResult = Box::pin(async move {
                    let msg = result?;
                    if let Some(mut replier) = replier {
                        let (envelope, rx) = Envelope::new_call(msg);
                        tx.send(envelope).await.map_err(|_| {
                            bterr!("failed to deliver message. Recipient may have halted.")
                        })?;
                        match rx.await {
                            Ok(reply) => {
                                let mut guard = buffer.lock().await;
                                guard.clear();
                                write_to(&reply, guard.deref_mut())?;
                                let wire_reply = WireReply::Ok(&guard);
                                replier.reply(wire_reply).await
                            }
                            Err(err) => replier.reply_err(err.to_string(), None).await,
                        }
                    } else {
                        tx.send(Envelope::new_send(act_name, msg))
                            .await
                            .map_err(|_| {
                                bterr!("failed to deliver message. Recipient may have halted.")
                            })
                    }
                });
                fut
            }
        };
        let handle = tokio::task::spawn(activator(self, rx, act_id));
        let actor_handle = ActorHandle::new(handle, tx, deliverer);
        guard.insert(act_id, actor_handle);
        act_name
    }

    /// Registers a service activation closure for [ServiceId]. An error is returned if the
    /// [ServiceId] has already been registered.
    pub async fn register<Msg, F>(&self, id: ServiceId, spawner: F) -> Result<ServiceName>
    where
        Msg: 'static + CallMsg,
        F: 'static
            + Send
            + Sync
            + Fn(&'static Runtime) -> Pin<Box<dyn Future<Output = Result<ActorName>>>>,
    {
        let mut guard = self.registry.write().await;
        match guard.entry(id.clone()) {
            hash_map::Entry::Vacant(entry) => {
                entry.insert(ServiceRecord::new(spawner));
                Ok(ServiceName::new(self.path().clone(), id.clone()))
            }
            hash_map::Entry::Occupied(_) => {
                log::info!("Updated registration for service '{id}'.");
                Ok(ServiceName::new(self.path().clone(), id))
            }
        }
    }

    pub async fn take_service(&self, id: &ServiceId) -> Option<ServiceRecord> {
        let mut registry = self.registry.write().await;
        registry.remove(id)
    }

    /// Returns the [ActorHandle] for the actor with the given name.
    ///
    /// If there is no such actor in this runtime then a [RuntimeError::BadActorName] error is
    /// returned.
    ///
    /// Note that the actor will be aborted when the given handle is dropped (unless it has already
    /// returned when the handle is dropped), and no further messages will be delivered to it by
    /// this runtime.
    pub async fn take(&self, name: &ActorName) -> Result<ActorHandle> {
        if name.path == self.path {
            let mut guard = self.handles.write().await;
            if let Some(handle) = guard.remove(&name.act_id) {
                Ok(handle)
            } else {
                Err(RuntimeError::BadActorName(name.clone()).into())
            }
        } else {
            Err(RuntimeError::BadActorName(name.clone()).into())
        }
    }

    /// Returns the name of the actor in this runtime with the given actor ID.
    pub fn actor_name(&self, act_id: Uuid) -> ActorName {
        ActorName::new(self.path.clone(), act_id)
    }
}

impl Drop for Runtime {
    fn drop(&mut self) {
        panic!("A Runtime was dropped. Panicking to avoid undefined behavior.");
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum RuntimeError {
    BadActorName(ActorName),
}

impl Display for RuntimeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::BadActorName(name) => write!(f, "bad actor name: {name}"),
        }
    }
}

impl std::error::Error for RuntimeError {}

/// Closure type used to spawn new service providers.
pub type Spawner =
    Box<dyn Send + Sync + Fn(&'static Runtime) -> Pin<Box<dyn Future<Output = Result<ActorName>>>>>;

pub struct ServiceRecord {
    spawner: Spawner,
    actor_ids: Vec<Uuid>,
}

impl ServiceRecord {
    fn new<F>(spawner: F) -> Self
    where
        F: 'static
            + Send
            + Sync
            + Fn(&'static Runtime) -> Pin<Box<dyn Future<Output = Result<ActorName>>>>,
    {
        Self {
            spawner: Box::new(spawner),
            actor_ids: Vec::new(),
        }
    }
}

/// Represents the terminal state of an actor, where it stops processing messages and halts.
pub struct End;

impl End {
    /// Returns the identifier for this type which is expected in protocol definitions.
    pub fn ident() -> &'static str {
        stringify!(End)
    }
}

#[allow(dead_code)]
/// Delivered to an actor implementation when it starts up.
pub struct Activate {
    /// A reference to the `Runtime` which is running this actor.
    rt: &'static Runtime,
    /// The ID assigned to this actor.
    act_id: Uuid,
}

impl Activate {
    pub fn new(rt: &'static Runtime, act_id: Uuid) -> Self {
        Self { rt, act_id }
    }
}

/// Deserializes replies sent over the wire.
pub struct ReplyCallback<T> {
    _phantom: PhantomData<T>,
}

impl<T: CallMsg> ReplyCallback<T> {
    pub fn new() -> Self {
        Self {
            _phantom: PhantomData,
        }
    }
}

impl<T: CallMsg> Default for ReplyCallback<T> {
    fn default() -> Self {
        Self::new()
    }
}

impl<T: CallMsg> DeserCallback for ReplyCallback<T> {
    type Arg<'de> = WireReply<'de> where T: 'de;
    type Return = Result<T::Reply>;
    type CallFut<'de> = Ready<Self::Return> where T: 'de, T::Reply: 'de;
    fn call<'de>(&'de mut self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
        let result = match arg {
            WireReply::Ok(slice) => from_slice(slice).map_err(|err| err.into()),
            WireReply::Err(msg) => Err(StringError::new(msg.to_string()).into()),
        };
        ready(result)
    }
}

struct SendReplyCallback {
    replier: Option<Replier>,
}

impl SendReplyCallback {
    fn new(replier: Replier) -> Self {
        Self {
            replier: Some(replier),
        }
    }
}

impl DeserCallback for SendReplyCallback {
    type Arg<'de> = WireReply<'de>;
    type Return = Result<()>;
    type CallFut<'de> = impl 'de + Future<Output = Self::Return>;
    fn call<'de>(&'de mut self, arg: Self::Arg<'de>) -> Self::CallFut<'de> {
        async move {
            if let Some(mut replier) = self.replier.take() {
                replier.reply(arg).await
            } else {
                Ok(())
            }
        }
    }
}

/// This struct implements the server callback for network messages.
#[derive(Clone)]
struct RuntimeCallback {
    rt: &'static Runtime,
}

impl RuntimeCallback {
    fn new(rt: &'static Runtime) -> Self {
        Self { rt }
    }

    async fn deliver_local(&self, msg: WireMsg<'_>, replier: Option<Replier>) -> Result<()> {
        let guard = self.rt.handles.read().await;
        if let Some(handle) = guard.get(&msg.to.act_id) {
            let envelope = if let Some(replier) = replier {
                WireEnvelope::Call { msg, replier }
            } else {
                WireEnvelope::Send { msg }
            };
            (handle.deliverer)(envelope).await
        } else {
            Err(bterr!("invalid actor name: {}", msg.to))
        }
    }

    async fn route_msg(&self, msg: WireMsg<'_>, replier: Option<Replier>) -> Result<()> {
        let guard = self.rt.peers.read().await;
        if let Some(tx) = guard.get(msg.to.path()) {
            if let Some(replier) = replier {
                let callback = SendReplyCallback::new(replier);
                tx.call(msg, callback).await?
            } else {
                tx.send(msg).await
            }
        } else {
            Err(bterr!(
                "unable to deliver message to peer at '{}'",
                msg.to.path()
            ))
        }
    }
}

impl MsgCallback for RuntimeCallback {
    type Arg<'de> = WireMsg<'de>;
    type CallFut<'de> = impl 'de + Future<Output = Result<()>>;
    fn call<'de>(&'de self, arg: bttp::MsgReceived<Self::Arg<'de>>) -> Self::CallFut<'de> {
        async move {
            let (_, body, replier) = arg.into_parts();
            if body.to.path() == self.rt.path() {
                self.deliver_local(body, replier).await
            } else {
                self.route_msg(body, replier).await
            }
        }
    }
}

/// A unique identifier for a particular service.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ServiceId(#[serde(with = "smart_ptr")] Arc<String>);

impl ServiceId {
    pub fn new(value: Arc<String>) -> Self {
        Self(value)
    }
}

impl AsRef<str> for ServiceId {
    fn as_ref(&self) -> &str {
        self.0.as_str()
    }
}

impl<T: Into<String>> From<T> for ServiceId {
    fn from(value: T) -> Self {
        Self(Arc::new(value.into()))
    }
}

impl Display for ServiceId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.as_ref())
    }
}

/// A unique identifier for a service.
///
/// A service is a collection of actors in the same directory which provide some functionality.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ServiceName {
    /// The path to the directory containing the service.
    #[serde(with = "smart_ptr")]
    path: Arc<BlockPath>,
    /// The id of the service.
    service_id: ServiceId,
}

impl ServiceName {
    pub fn new(path: Arc<BlockPath>, service_id: ServiceId) -> Self {
        Self { path, service_id }
    }
}

/// Indicates the set of service providers a message is addressed to.
///
/// The message could be delivered to any of the service providers in this set, at random.
pub struct ServiceAddr {
    /// The [ServiceName] to address the message to.
    name: ServiceName,
    #[allow(dead_code)]
    /// Indicates if the message should be routed towards the root of the tree or away from it.
    rootward: bool,
}

impl ServiceAddr {
    pub fn new(name: ServiceName, rootward: bool) -> Self {
        Self { name, rootward }
    }

    pub fn path(&self) -> &Arc<BlockPath> {
        &self.name.path
    }

    pub fn service_id(&self) -> &ServiceId {
        &self.name.service_id
    }
}

/// A unique identifier for a specific actor activation.
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone, Serialize, Deserialize)]
pub struct ActorName {
    /// The path to the directory containing this actor.
    #[serde(with = "smart_ptr")]
    path: Arc<BlockPath>,
    /// A unique identifier for an actor activation. Even as an actor transitions to different types
    /// as it handles messages, this value does not change. Thus this value can be used to trace an
    /// actor through a series of state transitions.
    act_id: Uuid,
}

impl ActorName {
    pub fn new(path: Arc<BlockPath>, act_id: Uuid) -> Self {
        Self { path, act_id }
    }

    pub fn path(&self) -> &Arc<BlockPath> {
        &self.path
    }

    pub fn act_id(&self) -> Uuid {
        self.act_id
    }
}

impl Display for ActorName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}@{}", self.act_id, self.path)
    }
}

/// Trait for messages which expect exactly one reply.
pub trait CallMsg: Serialize + DeserializeOwned + Send + Sync {
    /// The reply type expected for this message.
    type Reply: Serialize + DeserializeOwned + Send + Sync;
}

/// Trait for messages which expect exactly zero replies.
pub trait SendMsg: CallMsg {}

/// A type used to express when a reply is not expected for a message type.
#[derive(Serialize, Deserialize)]
pub enum NoReply {}

/// The maximum number of messages which can be kept in an actor's mailbox.
const MAILBOX_LIMIT: usize = 32;

/// The type of messages sent over the wire between runtimes.
#[derive(Serialize, Deserialize)]
pub struct WireMsg<'a> {
    to: ActorName,
    from: ActorName,
    payload: &'a [u8],
}

impl<'a> WireMsg<'a> {
    pub fn new(to: ActorName, from: ActorName, payload: &'a [u8]) -> Self {
        Self { to, from, payload }
    }
}

impl<'a> bttp::CallMsg<'a> for WireMsg<'a> {
    type Reply<'r> = WireReply<'r>;
}

impl<'a> bttp::SendMsg<'a> for WireMsg<'a> {}

#[derive(Serialize, Deserialize)]
pub enum WireReply<'a> {
    Ok(&'a [u8]),
    Err(&'a str),
}

/// A wrapper around [WireMsg] which indicates whether a call or send was executed.
enum WireEnvelope<'de> {
    Send { msg: WireMsg<'de> },
    Call { msg: WireMsg<'de>, replier: Replier },
}

impl<'de> WireEnvelope<'de> {
    fn into_parts(self) -> (WireMsg<'de>, Option<Replier>) {
        match self {
            Self::Send { msg } => (msg, None),
            Self::Call { msg, replier } => (msg, Some(replier)),
        }
    }
}

pub enum EnvelopeKind<T: CallMsg> {
    Call {
        reply: Option<oneshot::Sender<T::Reply>>,
    },
    Send {
        from: ActorName,
    },
}

impl<T: CallMsg> EnvelopeKind<T> {
    pub fn name(&self) -> &'static str {
        match self {
            Self::Call { .. } => "Call",
            Self::Send { .. } => "Send",
        }
    }
}

/// Wrapper around a message type `T` which indicates who the message is from and, if the message
/// was dispatched with `call`, provides a channel to reply to it.
pub struct Envelope<T: CallMsg> {
    msg: T,
    kind: EnvelopeKind<T>,
}

impl<T: CallMsg> Envelope<T> {
    pub fn new(msg: T, kind: EnvelopeKind<T>) -> Self {
        Self { msg, kind }
    }

    /// Creates a new envelope containing the given message which does not expect a reply.
    fn new_send(from: ActorName, msg: T) -> Self {
        Self {
            kind: EnvelopeKind::Send { from },
            msg,
        }
    }

    /// Creates a new envelope containing the given message which expects exactly one reply.
    fn new_call(msg: T) -> (Self, oneshot::Receiver<T::Reply>) {
        let (tx, rx) = oneshot::channel::<T::Reply>();
        let envelope = Self {
            kind: EnvelopeKind::Call { reply: Some(tx) },
            msg,
        };
        (envelope, rx)
    }

    /// Returns the name of the actor which sent this message.
    pub fn from(&self) -> Option<&ActorName> {
        match &self.kind {
            EnvelopeKind::Send { from } => Some(from),
            _ => None,
        }
    }

    /// Returns a reference to the message in this envelope.
    pub fn msg(&self) -> &T {
        &self.msg
    }

    /// Sends a reply to this message.
    ///
    /// If this message is not expecting a reply, or if this message has already been replied to,
    /// then an error is returned.
    pub fn reply(&mut self, reply: T::Reply) -> Result<()> {
        match &mut self.kind {
            EnvelopeKind::Call { reply: tx } => {
                if let Some(tx) = tx.take() {
                    tx.send(reply).map_err(|_| bterr!("Failed to send reply."))
                } else {
                    Err(bterr!("Reply has already been sent."))
                }
            }
            _ => Err(bterr!("Can't reply to '{}' messages.", self.kind.name())),
        }
    }

    /// Returns true if this message expects a reply and it has not already been replied to.
    pub fn needs_reply(&self) -> bool {
        matches!(&self.kind, EnvelopeKind::Call { .. })
    }

    pub fn split(self) -> (T, EnvelopeKind<T>) {
        (self.msg, self.kind)
    }
}

type FutureResult = Pin<Box<dyn Send + Future<Output = Result<()>>>>;

pub struct ActorHandle {
    handle: Option<JoinHandle<()>>,
    sender: Box<dyn Send + Sync + Any>,
    deliverer: Box<dyn Send + Sync + Fn(WireEnvelope<'_>) -> FutureResult>,
}

impl ActorHandle {
    fn new<T, F>(handle: JoinHandle<()>, sender: mpsc::Sender<Envelope<T>>, deliverer: F) -> Self
    where
        T: 'static + CallMsg,
        F: 'static + Send + Sync + Fn(WireEnvelope<'_>) -> FutureResult,
    {
        Self {
            handle: Some(handle),
            sender: Box::new(sender),
            deliverer: Box::new(deliverer),
        }
    }

    fn sender<T: 'static + CallMsg>(&self) -> Result<&mpsc::Sender<Envelope<T>>> {
        self.sender
            .downcast_ref::<mpsc::Sender<Envelope<T>>>()
            .ok_or_else(|| bterr!("Attempt to send message as the wrong type."))
    }

    /// Sends a message to the actor represented by this handle.
    pub async fn send<T: 'static + SendMsg>(&self, from: ActorName, msg: T) -> Result<()> {
        let sender = self.sender()?;
        sender
            .send(Envelope::new_send(from, msg))
            .await
            .map_err(|_| bterr!("failed to enqueue message"))?;
        Ok(())
    }

    pub async fn call_through<T: 'static + CallMsg>(&self, msg: T) -> Result<T::Reply> {
        let sender = self.sender()?;
        let (envelope, rx) = Envelope::new_call(msg);
        sender
            .send(envelope)
            .await
            .map_err(|_| bterr!("failed to enqueue call"))?;
        let reply = rx.await?;
        Ok(reply)
    }

    pub async fn returned(&mut self) -> Result<()> {
        if let Some(handle) = self.handle.take() {
            handle.await?;
        }
        Ok(())
    }

    pub fn abort(&mut self) {
        if let Some(handle) = self.handle.take() {
            handle.abort();
        }
    }
}

impl Drop for ActorHandle {
    fn drop(&mut self) {
        self.abort();
    }
}
